/***************************************************************************
 *   Copyright (C) 2007 by Saritha Kalyanam   				   *
 *   sari.kaly@gmail.com   						   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
extern int errno;

#include "stun.h"
#include "types.h"
#include "debug.h"
                 
extern int h_errno;

const char * stun_server_list[] = {
    "stunserver.org",
    "stun.fwdnet.net",
    "stun01.sipphone.com",
    "stun.softijoys.com",
    "stun.voipbuster.com",
    "stun.voxgratia.org",
    "stun.xten.com",
    "stun.ekiga.net",
    0
};


#if 0
static const char *
stun_msg2str(int msg)
{
    switch(msg) {
        case STUN_BINDREQ:
            return "Binding Request";
        case STUN_BINDRESP:
            return "Binding Response";
        case STUN_BINDERR:
            return "Binding Error Response";
        case STUN_SECREQ:
            return "Shared Secret Request";
        case STUN_SECRESP:
            return "Shared Secret Response";
        case STUN_SECERR:
            return "Shared Secret Error Response";
    }
    return "Non-RFC3489 Message";
}

static const char *
stun_attr2str(int msg)
{
    switch(msg) {
        case STUN_MAPPED_ADDRESS:
            return "Mapped Address";
        case STUN_RESPONSE_ADDRESS:
            return "Response Address";
        case STUN_CHANGE_REQUEST:
            return "Change Request";
        case STUN_SOURCE_ADDRESS:
            return "Source Address";
        case STUN_CHANGED_ADDRESS:
            return "Changed Address";
        case STUN_USERNAME:
            return "Username";
        case STUN_PASSWORD:
            return "Password";
        case STUN_MESSAGE_INTEGRITY:
            return "Message Integrity";
        case STUN_ERROR_CODE:
            return "Error Code";
        case STUN_UNKNOWN_ATTRIBUTES:
            return "Unknown Attributes";
        case STUN_REFLECTED_FROM:
            return "Reflected From";
    }
    return "Non-RFC3489 Attribute";
}
#endif

struct stun_state {
    const char *username;
    const char *password;
    struct stun_addr *mapped_addr;
    struct stun_addr *changed_addr;
};

static long int
stun_random(void)
{
    return random();
}

static int 
stun_process_attr(struct stun_state *state, struct stun_attr *attr)
{

#if 0
    if (stundebug)
        ast_verbose("Found STUN Attribute %s (%04x), length %d\n",
                    stun_attr2str(ntohs(attr->attr)), 
                    ntohs(attr->attr), ntohs(attr->len));
#endif

    switch(ntohs(attr->attr)) {
        case STUN_USERNAME:
            state->username = (const char *) (attr->value);
            break;
        case STUN_PASSWORD:
            state->password = (const char *) (attr->value);
            break;
        case STUN_MAPPED_ADDRESS:
            state->mapped_addr = (struct stun_addr *) (attr->value);
            state->mapped_addr->addr = state->mapped_addr->addr;
            state->mapped_addr->port = state->mapped_addr->port;
            break;
        case STUN_CHANGED_ADDRESS:
            state->changed_addr = (struct stun_addr *) (attr->value);
            state->changed_addr->addr = state->changed_addr->addr;
            state->changed_addr->port = state->changed_addr->port;
            break;
        default:
            break;
#if 0
            if (stundebug)
                ast_verbose("Ignoring STUN attribute %s (%04x), length %d\n",
                            stun_attr2str(ntohs(attr->attr)), ntohs(attr->attr), ntohs(attr->len));
#endif
    }
    return 0;
}

static void 
append_attr_string(struct stun_attr **attr, int attrval, 
                    const char *s, int *len, int *left)
{
    int size = sizeof(**attr) + strlen(s);
    if (*left > size) {
        (*attr)->attr = htons(attrval);
        (*attr)->len = htons(strlen(s));
        memcpy((*attr)->value, s, strlen(s));
        (*attr) = (struct stun_attr *)((*attr)->value + strlen(s));
        *len += size;
        *left -= size;
    }
}

#if 0
static void 
append_attr_address(struct stun_attr **attr, int attrval, 
                        struct sockaddr_in *sin, int *len, int *left)
{
    int size = sizeof(**attr) + 8;
    struct stun_addr *addr;
    if (*left > size) {
        (*attr)->attr = htons(attrval);
        (*attr)->len = htons(8);
        addr = (struct stun_addr *)((*attr)->value);
        addr->unused = 0;
        addr->family = 0x01;
        addr->port = sin->sin_port;
        addr->addr = sin->sin_addr.s_addr;
        (*attr) = (struct stun_attr *)((*attr)->value + 8);
        *len += size;
        *left -= size;
    }
}
#endif

static int 
stun_send(int s, struct sockaddr_in *dst, struct stun_header *resp)
{
    return sendto(s, resp, ntohs(resp->msglen) + sizeof(*resp), 0,
                  (struct sockaddr *)dst, sizeof(*dst));
}

static void 
stun_req_id(struct stun_header *req)
{
    int x;
    for (x=0;x<4;x++)
        req->id.id[x] = stun_random();
}

void
stun_request(int sock, struct sockaddr_in *suggestion,
                        const char *username)
{
    struct stun_header *req;
    unsigned char reqdata[1024];
    int reqlen, reqleft;
    struct stun_attr *attr;
    int ret;

    req = (struct stun_header *)reqdata;
    stun_req_id(req);
    reqlen = 0;
    reqleft = sizeof(reqdata) - sizeof(struct stun_header);
    req->msgtype = 0;
    req->msglen = 0;
    attr = (struct stun_attr *)req->ies;
    if (username)
        append_attr_string(&attr, STUN_USERNAME, username, &reqlen, &reqleft);
    req->msglen = htons(reqlen);
    req->msgtype = htons(STUN_BINDREQ);
    ret = stun_send(sock, suggestion, req);
    printf("stun_send %d bytes\n", ret);
    ASSERT(ret >= 0);
}


int 
stun_get_mapped_addr(unsigned char *data, int len, 
                            struct sockaddr_in *sin)
{
    struct stun_header *hdr = (struct stun_header *)data;
    struct stun_attr *attr;
    struct stun_state st;
    int ret = STUN_IGNORE;

    bzero(&st, sizeof(st));

    if (len < sizeof(struct stun_header)) {
#if 0
        if (option_debug)
            ast_log(LOG_DEBUG, "Runt STUN packet (only %zd, wanting at least %zd)\n", len, sizeof(struct stun_header));
#endif
        return -1;
    }
#if 0
    if (stundebug)
        ast_verbose("STUN Packet, msg %s (%04x), length: %d\n", stun_msg2str(ntohs(hdr->msgtype)), ntohs(hdr->msgtype), ntohs(hdr->msglen));
#endif
    if (ntohs(hdr->msglen) > len - sizeof(struct stun_header)) {
#if 0
        if (option_debug)
            ast_log(LOG_DEBUG, "Scrambled STUN packet length (got %d, expecting %zd)\n", ntohs(hdr->msglen), len - sizeof(struct stun_header));
#endif
    } else
        len = ntohs(hdr->msglen);
    data += sizeof(struct stun_header);
    memset(&st, 0, sizeof(st));
    while(len > 0) {
        if (len < sizeof(struct stun_attr)) {
#if 0
            if (option_debug)
                ast_log(LOG_DEBUG, "Runt Attribute (got %zd, expecting %zd)\n", len, sizeof(struct stun_attr));
#endif
            break;
        }
        attr = (struct stun_attr *)data;
        if (ntohs(attr->len) > len) {
#if 0
            if (option_debug)
                ast_log(LOG_DEBUG, "Inconsistent Attribute (length %d exceeds remaining msg len %zd)\n", ntohs(attr->len), len);
#endif
            break;
        }
        if (stun_process_attr(&st, attr)) {
#if 0
            if (option_debug)
                ast_log(LOG_DEBUG, "Failed to handle attribute %s (%04x)\n", stun_attr2str(ntohs(attr->attr)), ntohs(attr->attr));
#endif
            break;
        }

        /* Clear attribute in case previous entry was a string */
        attr->attr = 0;
        data += ntohs(attr->len) + sizeof(struct stun_attr);
        len -= ntohs(attr->len) + sizeof(struct stun_attr);
    }

    if (len < 0) {
        ret = STUN_IGNORE;
        return ret;
    }

    /* we have parsed everything correctly */
    if (st.mapped_addr) {
        bzero(sin, sizeof(struct sockaddr_in));
        sin->sin_family = st.mapped_addr->family;
        sin->sin_port = st.mapped_addr->port;
        memcpy(&sin->sin_addr.s_addr, &st.mapped_addr->addr, 
                sizeof(struct in_addr));
    }

    /* we may need to send another stun pkt */
    if (st.changed_addr) {
    }

    return ret;
}

/* FIXME: we are doing only the first of three STUN tests */
int
stun_get_ext_ip_addr(struct sockaddr_in *local_addr4, 
                        struct sockaddr_in *ext_addr4)
{
    int stun_sock = 0;
    struct sockaddr_in stun_addr;
    struct hostent *stun_he = NULL;
    struct sockaddr_in from;
    socklen_t fromlen;
    u8 buf[2048];
    int timeout;
    bool reply = FALSE;
    int i = 0;
    int ret = FAILURE;

    stun_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (stun_sock < 0) {
        return FAILURE;
    }

    bzero(&stun_addr, sizeof(struct sockaddr_in));
    memcpy(&stun_addr, local_addr4, sizeof(struct sockaddr_in));

    ret = bind(stun_sock, (struct sockaddr *)&stun_addr, sizeof(stun_addr));
    if (ret < 0) {
        close(stun_sock);
        return FAILURE;
    }

    /* FIXME: randomize the stun server list */

    for (i = 0; stun_server_list[i]; i++) {

        stun_he = gethostbyname(stun_server_list[i]);
        if (!stun_he)
            continue;

        bzero(&stun_addr, sizeof(struct sockaddr_in));
        stun_addr.sin_family = stun_he->h_addrtype;
        stun_addr.sin_port = htons(STUN_PORT);
        memcpy(&stun_addr.sin_addr.s_addr, stun_he->h_addr, sizeof(in_addr_t));

        /* send the stun request */
        stun_request(stun_sock, &stun_addr, NULL);
        printf("%s STUN request sent\n", stun_server_list[i]);      

        timeout = MAX_STUN_TIMEOUT;
        while (timeout) {
            errno = 0;
            ret = recvfrom(stun_sock, buf, sizeof(buf), MSG_DONTWAIT,
                           (struct sockaddr *)&from, &fromlen);
            if (ret < 0) {
                if (errno != EAGAIN) {
                    break;
                } else {
                    timeout--;
                    sleep(1);
                    continue;
                } 
            } else if (ret == 0) {
                break;
            } else {
                stun_get_mapped_addr(buf, ret, ext_addr4);
                
                printf("%s STUN reply received\n", stun_server_list[i]);
                reply = TRUE;
                ret = SUCCESS;
                break;
            }
        }
        
        if (reply) {
            printf("%s STUN reply success\n", stun_server_list[i]);         
            break;
        }
    }

    close(stun_sock);

    return ret;
}
